import boto3
import click
import os
import time
import yaml

from botocore.exceptions import NoCredentialsError
from os.path import join, expanduser

from .click_helpers import info, unimportant, panic, error

def read_config(ctx):
  ctx.obj['config'].seek(0)
  config = yaml.safe_load(ctx.obj['config'])
  config = {} if config is None else config
  return config

def write_config(ctx, config):
  ctx.obj['config'].seek(0)
  ctx.obj['config'].truncate(0)
  yaml.dump(config, ctx.obj['config'])

@click.group()
@click.option('-c', '--config', type=click.File(mode='a+'), default=expanduser(join("~", ".nf-batch-runner")))
@click.option('--region', default=None)
@click.pass_context
def cli(ctx, config, region):
  """
  Run Nextflow in AWS Batch

  Helps you to setup and manage Nextflow environments in AWS Batch
  """

  ctx.ensure_object(dict)
  ctx.obj['config'] = config

  try:
    sts = boto3.client("sts")
    ctx.obj['account_details'] = sts.get_caller_identity()
  except NoCredentialsError:
    panic("No AWS credentials found")

  config = read_config(ctx)
  if config.get('region') and region:
    panic(f"Region is already set in {config.name}, no need to set it again")
  elif not region:
    region = "eu-west-2"
  elif config.get('region'):
    region = config.get('region')

  boto3.setup_default_session(region_name=region)
 
  ctx.obj['region'] = region
  
@cli.command()
@click.option('--yes', '-y', default=False, is_flag=True)
@click.pass_context
def delete(ctx, yes):
  """Delete your enviroment"""
  from .aws import delete_bucket, delete_stack
  config = read_config(ctx)
  
  environment = config['environment']
  stack = config['stack_id']
  
  if not yes:
    click.confirm(f"Are you sure you want to delete the environment '{environment}'", abort=True)

  outputs = config.get("outputs", {})
  try:
    files_bucket = outputs["FilesBucket"]
  except KeyError:
    pass
  else:
    if not yes:
      delete_files = click.confirm(f"Do you want to delete the files in {files_bucket}")
    else:
      delete_files = True

    if delete_files:
      n_objects = delete_bucket(files_bucket)
      unimportant(f"Deleted {n_objects} objects from {files_bucket}")

  try:
    frontend_bucket = outputs["FrontendBucket"]
  except KeyError:
    pass
  else:
    delete_bucket(frontend_bucket)
    unimportant(f"Deleted {frontend_bucket}")

  delete_stack(stack)
  info(f"Deleted {environment}")

  config_path = ctx.obj['config'].name
  os.rename(config_path, f"{config_path}.{int(time.time())}.backup")

@cli.command()
def user():
  """Add, remove, update users of the frontend"""
  # Ask for a name
  # Ask for a password
  # Lookup the UsersDb
  # Add or update the User
  raise NotImplementedError()

@cli.command()
@click.option('--yes', '-y', default=False, is_flag=True)
@click.option('--ami', default=None, type=str)
@click.argument('environment', default='')
@click.pass_context
def create(ctx, yes, ami, environment):
  """Create your enviroment"""
  from .environment import pick_password, pick_environment, lookup_version, fetch_template, pick_ami
  from .aws import create_stack, wait_for_resource, wait_for_stack, sync, create_user

  config = read_config(ctx)
  config['region'] = ctx.obj['region']

  if ami is not None:
    config['AMI'] = ami
  elif config['region'] != 'eu-west-2':
    error("We currently only have an AMI built for eu-west-2 (London).  You may be able to use this tool in other regions if you build your own AMI")
    config['AMI'] = pick_ami()

  if config.get('environment'):
    panic(f"You already have an environment called {config.get('environment')}")
  config['environment'] = pick_environment(environment)
  password = pick_password()

  account_details = ctx.obj['account_details']
  if not yes:
    click.confirm(f"Are you happy to install nf-batch-runner in account {account_details['Account']}", abort=True)
    if not click.confirm(f"Does you user ({account_details['Arn'].partition('/')[-1]}) have administrator privileges"):
      panic("It is unlikely that we could create all the required resources, aborting")
    frontend = click.confirm("Do you want to deploy the web interface", default=True)
  else:
    frontend = True
  config['version'] = lookup_version(config)

  template = fetch_template(config, config['version'])

  config['stack_id'] = create_stack(config['environment'], template, frontend, config['version']['version'], config.get('AMI'))

  write_config(ctx, config)
  if frontend:
    unimportant("Waiting for the S3 bucket for the frontend (takes around a minute)", nl=False)
    bucket_name, ok, message = wait_for_resource(config['stack_id'], 'FrontendBucket')
    if not ok:
      error(" ✗")
      error("There was a problem creating the bucket")
      panic(message)
    unimportant(" ✓")
    sync(config['version']['frontend'], f"s3://{bucket_name}/frontend/{config['version']['version']}/")

  unimportant("Waiting for the users database (takes around a minute)", nl=False)
  users_db, ok, message = wait_for_resource(config['stack_id'], 'UsersDb')
  if not ok:
    error(" ✗")
    error("There was a problem creating the users database")
    panic(message)
  unimportant(" ✓")
  
  unimportant("Creating the users in the database", nl=False)
  create_user(users_db, "__GLOBAL", credits=1000)
  create_user(users_db, "admin", password, credits=100)
  unimportant(" ✓")

  outputs = wait_for_stack(config['stack_id'])
  config['outputs'] = outputs

  if frontend:
    info(f"Visit https://{outputs['FrontendUrl']} and login with the user 'admin'")
  else:
    info(f"Your environment is not ready")
  info(f"You can upload your files here: https://s3.console.aws.amazon.com/s3/buckets/{outputs['FilesBucket']}/")

  write_config(ctx, config)



import click
import time

from contextlib import contextmanager

def info(message, **vargs):
  click.echo(message, **vargs)

def unimportant(message, **vargs):
  click.secho(message, fg='cyan', **vargs)

def error(message, **vargs):
  click.secho(message, fg='red')

def panic(message, status=1, **vargs):
  error(message, **vargs)
  exit(status)

@contextmanager
def fake_progress(eta, half_life, *args, **vargs):
  """
  Retuns an approximate progress bar

  Progress is 90% after 'eta' seconds.  The remaining progress halves
  every 'half_life' seconds.
  """
  vargs['length'] = 100
  with click.progressbar(*args, **vargs) as bar:
    previous_progress = 0
    started = time.time()
    def tick():
      nonlocal previous_progress
      elapsed = time.time() - started
      if elapsed < eta:
        progress = 90 * elapsed / eta
      else:
        overdue = elapsed - eta
        progress = 100 - 10 * (0.5 ** (overdue / half_life))
      
      bar.update(progress - previous_progress)
      previous_progress = progress
    
    yield tick


import boto3
import click
import json
import re

from botocore.exceptions import ClientError

from .click_helpers import panic, error, unimportant

def pick_environment(environment, attempts=3):
  if attempts <= 0:
    panic("Giving up")
  elif environment == '':
    click.echo("What do you want to call your environment?")
    environment = click.prompt("Environment")
  elif not re.match("^[a-zA-Z][a-zA-Z0-9\-]*$", environment):
    error("Environment must start with a letter and only contain letters numbers and '-'")
    environment = click.prompt("Environment")
  elif len(environment) > 128:
    error("Environment must be less than 128 characters")
    environment = click.prompt("Environment")
  else:
    try:
      cf = boto3.client('cloudformation')
      cf.describe_stacks(StackName=environment)
    except ClientError:
      # stack doesn't exist
      return environment
    else:
      error("Environment with that name already exists")
      environment = click.prompt("Environment")
  return pick_environment(environment, attempts-1)

def pick_password(attempts=3):
  click.echo("Please give us a password you will use to login to the frontend")
  password = click.prompt("Password", default=None, hide_input=True)
  if password != click.prompt("Repeat", default=None, hide_input=True):
    error("Those passwords don't match")
    if attempts <= 0:
      panic("Giving up")
    return pick_password(attempts-1)
  return password

def pick_ami(attempts=3):
  click.echo("Which AMI would you like to use?")
  ami = click.prompt("AMI ID")
  if not ami:
    if attempts <= 0:
      panic("Giving up")
    return pick_ami(attempts-1)
  return ami

def lookup_version(config):
  from .aws import download

  unimportant("Fetching the latest version", nl=False)
  if config.get('version'):
    unimportant(f" ({config['version']['version']}) [from config]")
    return config['version']
  versions = json.load(download(config, 'versions.json')['Body'])
  unimportant(f" ({versions['latest']['version']})")
  return versions['latest']

def fetch_template(config, version):
  from .aws import download

  unimportant("Downloading the infrastructure template", nl=False)
  template = download(config, version['backend'])['Body'].read()
  unimportant(" ✓")
  return template.decode('utf8')


import base64
import bcrypt
import boto3
import click
import datetime
import os
import time
import typing

from botocore.exceptions import ClientError
from decimal import Decimal
from itertools import groupby

from .click_helpers import unimportant, fake_progress

Status = typing.NamedTuple(
  'Status', 
  [
    ('in_progress', bool), 
    ('failed', bool),
  ]
)

s3 = boto3.client('s3')
cf = boto3.client('cloudformation')

def parse_status(status):
  all_statuses = [
    'CREATE_IN_PROGRESS'
    'CREATE_FAILED'
    'CREATE_COMPLETE'
    'ROLLBACK_IN_PROGRESS'
    'ROLLBACK_FAILED'
    'ROLLBACK_COMPLETE'
    'DELETE_IN_PROGRESS'
    'DELETE_FAILED'
    'DELETE_COMPLETE'
    'UPDATE_IN_PROGRESS'
    'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS'
    'UPDATE_COMPLETE'
    'UPDATE_ROLLBACK_IN_PROGRESS'
    'UPDATE_ROLLBACK_FAILED'
    'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS'
    'UPDATE_ROLLBACK_COMPLETE'
    'REVIEW_IN_PROGRESS'
  ]
  return Status(
    status.endswith('_IN_PROGRESS'),
    'ROLLBACK_' in status or status.endswith('_FAILED'),
  )

def download(config, url):
  if url.startswith('s3://'):
    bucket, _, key = url[5:].partition('/')
  else:
    bucket = config.get("source_bucket", 'nf-batch-runner')
    key = url
  return s3.get_object(Bucket=bucket, Key=key)

def create_stack(environment, template, frontend, version, ami=None):
  session_secret = base64.b64encode(os.urandom(256//8)).decode('utf8')
  now = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
  params = dict(
    StackName=environment,
    TemplateBody=template,
    Parameters=[
        {
            'ParameterKey': 'DeployFrontend',
            'ParameterValue': 'yes' if frontend else 'no',
        },
        {
            'ParameterKey': 'SessionSecret',
            'ParameterValue': session_secret,
        },
        {
          'ParameterKey': 'FrontendStaticPrefix',
          'ParameterValue': f"/frontend/{version}",
        }
    ],
    TimeoutInMinutes=60,
    Capabilities=[
        'CAPABILITY_NAMED_IAM',
    ],
    OnFailure='DELETE',
    Tags=[
        {
            'Key': 'nf-batch-runner',
            'Value': environment
        },
        {
            'Key': 'nf-batch-runner-version',
            'Value': version
        },
    ]
  )
  if ami is not None:
    params['Parameters'].append(
      {
          'ParameterKey': 'WorkerNodeAMI',
          'ParameterValue': ami,
      }
    )
  response = cf.create_stack(**params)
  return response['StackId']

def wait_for_resource(stack, resource):
  while True:
    try:
      response = cf.describe_stack_resource(
        StackName=stack,
        LogicalResourceId=resource
      )
    except ClientError as e:
      time.sleep(5)
      continue
    raw_status = response['StackResourceDetail']['ResourceStatus']
    status = parse_status(raw_status)
    if status.failed:
      return response['StackResourceDetail']['PhysicalResourceId'], False, response['StackResourceDetail']['ResourceStatusReason']
    if status.in_progress:
      time.sleep(5)
      continue
    return response['StackResourceDetail']['PhysicalResourceId'], True, "Complete"

MINUTES = 60

def wait_for_stack(stack):
  expected_runtime = 20 * MINUTES
  progress_half_life = 5 * MINUTES
  unimportant("It can take 15-20 minutes to create your environment")
  with fake_progress(expected_runtime, progress_half_life) as tick:
    while True:
      tick()
      try:
        response = cf.describe_stacks(StackName=stack)
      except ClientError as e:
        time.sleep(5)
        continue
      stack_details = response['Stacks'][0]
      raw_status = stack_details['StackStatus']
      status = parse_status(raw_status)
      if status.failed:
        outputs = {}
        break
      if status.in_progress:
        time.sleep(5)
        continue
      outputs = stack_details['Outputs']
      outputs = { o['OutputKey']: o['OutputValue'] for o in outputs }
      break
  return outputs

def list_bucket_contents(bucket, prefix, ContinuationToken=None):
  params = dict(
    Bucket=bucket,
    MaxKeys=100,
    Prefix=prefix
  )
  if ContinuationToken:
    params['ContinuationToken'] = ContinuationToken
  resp = s3.list_objects_v2(**params)
  yield from resp.get('Contents', [])
  if resp['IsTruncated']:
    yield from list_bucket_contents(bucket, prefix, resp['NextContinuationToken'])

def sync(from_, to):
  from_bucket, _, from_prefix = from_[5:].partition('/')
  to_bucket, _, to_prefix = to[5:].partition('/')

  with click.progressbar(list(list_bucket_contents(from_bucket, from_prefix)), label=f"Syncing files from {from_bucket}") as bar:
    for obj in bar:
      from_key = obj['Key']
      to_key = from_key.replace(from_prefix, to_prefix)
      s3.copy_object(
        ACL='private',
        Bucket=to_bucket,
        CacheControl='max-age=0',
        CopySource={'Bucket': from_bucket, 'Key': from_key},
        Key=to_key,
        StorageClass='STANDARD'
      )

def create_user(users_db, user, password=None, credits=100):
  dynamodb = boto3.resource('dynamodb')
  table = dynamodb.Table(users_db)

  record = {
    "configs": [],
    "credits": {
      "balance": credits,
      "lastDebited": int(time.time()),
      "maxBalance": credits,
      "rate": Decimal(credits) / (28 * 24 * 60 * 60)
    },
    "metadata": {},
    "user": user
  }
  
  if password:
    record['password'] = bcrypt.hashpw(password.encode('utf8'), bcrypt.gensalt(14)).decode('utf8')

  table.put_item(
    Item=record,
    ReturnValues='NONE',
    ReturnConsumedCapacity='NONE',
    ReturnItemCollectionMetrics='NONE'
  )

def create_batches(objects, count):
  for _, batch in groupby(enumerate(objects), lambda el: el[0] // count):
    yield (el for _, el in batch)

def delete_objects(bucket, prefix):
  objects = list_bucket_contents(bucket, prefix)
  count = 0
  for batch in create_batches(objects, 1000):
    batch_objects = [ { 'Key': obj['Key'] } for obj in batch ]
    s3.delete_objects(
      Bucket=bucket,
      Delete={
          'Objects': batch_objects,
          'Quiet': True
      }
    )
    count += len(batch_objects)
  return count
    
def delete_bucket(bucket):
  try:
    n_objects = delete_objects(bucket, "")
    s3.delete_bucket(Bucket=bucket)
    return n_objects
  except s3.exceptions.NoSuchBucket:
    return 0
  
def delete_stack(stack):
  cf.delete_stack(
    StackName=stack,
  )



import setuptools

from os.path import dirname, join

with open(join(dirname(__file__), "README.md"), "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nf-batch-runner",
    version="0.1.12",
    author="Centre for Genomic Pathogen Surveillance",
    author_email="cgps@sanger.ac.uk",
    description="Installs nf-batch-runner in your AWS account",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/cgps/nf-batch-runner",
    packages=setuptools.find_packages(),
    py_modules=['nf-batch-runner'],
    entry_points='''
        [console_scripts]
        nf-batch-runner=nf_batch_runner.cli:cli
    ''',    
    install_requires=[
      "boto3>=1.9.227",
      "bcrypt>=3.1.7",
      "Click>=7.0",
      "PyYAML>=5.1.2"
    ],
    classifiers=[
      "Development Status :: 4 - Beta",
      "Programming Language :: Python :: 3",
      "Operating System :: OS Independent",
    ],
    keywords="aws batch nextflow bioinformatics pipeline",
    project_urls={
      'Source': 'https://gitlab.com/cgps/nf-batch-runner/',
    },
    python_requires='>=3',
)
